#include "gthb.h"

static char scratch[5][5];

static void clear_scratch() {
    int i, j;
    
    for (i = 0; i < 5; i++)
	for (j = 0; j < 5; j++)
	    scratch[i][j] = 0;
}

static void flood(struct board *b, enum who color, int x, int y) {
    /* off board */
    if (!(x >= 0 && x <= 4 && y >= 0 && y <= 4))
	return;
    /* already done */
    if (scratch[x][y])
	return;
    /* wrong color */
    if (b->cell[x][y] != color)
	return;
    /* ok */
    scratch[x][y] = 1;
    flood(b, color, x - 1, y);
    flood(b, color, x + 1, y);
    flood(b, color, x, y - 1);
    flood(b, color, x, y + 1);
}

static int group_border(int x, int y) {
    if (scratch[x][y])
	return 0;
    if (x > 0 && scratch[x - 1][y])
	return 1;
    if (x < 4 && scratch[x + 1][y])
	return 1;
    if (y > 0 && scratch[x][y - 1])
	return 1;
    if (y < 4 && scratch[x][y + 1])
	return 1;
    return 0;
}

int liberties(struct board *b, int x, int y) {
    int i, j;
    int n = 0;
    
    clear_scratch();
    flood(b, b->cell[x][y], x, y);
    for (i = 0; i < 5; i++)
	for (j = 0; j < 5; j++)
	    if (b->cell[i][j] == WHO_NONE && group_border(i, j))
		n++;
    return n;
}

int legal_move(struct board *b, int x, int y, enum who w) {
    int n;
    
    if (b->cell[x][y] != WHO_NONE)
	return 0;
    b->cell[x][y] = w;
    n = liberties(b, x, y);
    b->cell[x][y] = WHO_NONE;
    return (n > 0);
}

static enum who opponent(enum who w) {
  switch (w) {
  case WHO_BLACK:
    return WHO_WHITE;
  case WHO_WHITE:
    return WHO_BLACK;
  case WHO_NONE:
    break;
  }
  abort();
  /*NOTREACHED*/
}

void capture(struct board *b, int x, int y, enum who w) {
    int i, j, n;

    if (!(x >= 0 && x <= 4))
	return;
    if (!(y >= 0 && y <= 4))
	return;
    if (b->cell[x][y] != opponent(w))
	return;
    n = liberties(b, x, y);
    if (n > 0)
	return;
    for (i = 0; i < 5; i++)
	for (j = 0; j < 5; j++)
	    if (scratch[i][j])
		b->cell[i][j] = w;
}

void captures(struct board *b, int x, int y, enum who w) {
    capture(b, x - 1, y, w);
    capture(b, x + 1, y, w);
    capture(b, x, y + 1, w);
    capture(b, x, y - 1, w);
}

enum win referee(struct board *b) {
    int i, j;
    int black = 0;
    int white = 0;
    int legal = 0;
    
    if (currentboard.passes < 2)
	return WIN_NOTYET;
    for (i = 0; i < 5; i++) {
	for (j = 0; j < 5; j++) {
	    if (b->cell[i][j] == WHO_BLACK)
		black++;
	    else if (b->cell[i][j] == WHO_WHITE)
		white++;
	    else if (legal_move(b, i, j, WHO_BLACK))
		legal++;
	    else if (legal_move(b, i, j, WHO_WHITE))
		legal++;
	}
    }
    if (black > white)
	return WIN_BLACK;
    if (white > black)
	return WIN_WHITE;
    return WIN_DRAW;
}
