#include <Xm/Xm.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Form.h>
#include <Xm/Scale.h>
#include "inspmc.h"
#include <stdio.h>
#include "gthb.h"
#include "gthclient.h"

extern String fallback[];

static Widget cellwidget[25];
static Widget status;
static XtAppContext app;
static Pixmap black_stone_pixmap, white_stone_pixmap, blank_pixmap;
static XmString compWinString, humanWinString, drawString, emptyString;

static struct globalOptions {
  Pixmap black_stone_pixmap, white_stone_pixmap, blank_pixmap;
  String compWinString, humanWinString, drawString;
  Boolean compFirst;
  String compHost;
  int compServer;
} globalData;

static XtResource resources[] = {
  {"blackStonePixmap", "MovePixmap", XtRPixmap, sizeof(Pixmap),
   XtOffset(struct globalOptions *, black_stone_pixmap),
   XtRString, 0},

  {"whiteStonePixmap", "MovePixmap", XtRPixmap, sizeof(Pixmap),
   XtOffset(struct globalOptions *, white_stone_pixmap),
   XtRString, 0},

  {"blankPixmap", "MovePixmap", XtRPixmap, sizeof(Pixmap),
   XtOffset(struct globalOptions *, blank_pixmap),
   XtRString, 0},

  {"compWinString", "WinString", XtRString, sizeof(String),
   XtOffset(struct globalOptions *, compWinString),
   XtRString, (XtPointer) ""},

  {"humanWinString", "WinString", XtRString, sizeof(String),
   XtOffset(struct globalOptions *, humanWinString),
   XtRString, (XtPointer) ""},

  {"drawString", "WinString", XtRString, sizeof(String),
   XtOffset(struct globalOptions *, drawString),
   XtRString, (XtPointer) ""},

  {"compFirst", "CompFirst", XtRBoolean, sizeof(Boolean),
   XtOffset(struct globalOptions *, compFirst),
   XtRImmediate, (XtPointer) FALSE},

  {"compHost", "CompHost", XtRString, sizeof(String),
   XtOffset(struct globalOptions *, compHost),
   XtRString, (XtPointer) ""},

  {"compServer", "CompServer", XtRInt, sizeof(int),
   XtOffset(struct globalOptions *, compServer),
   XtRImmediate, (XtPointer) 0},
};

XrmOptionDescRec options[] = {
  { "-b", "*compFirst", XrmoptionNoArg, "false" },
  { "-w", "*compFirst", XrmoptionNoArg, "true" },
  { "-h", "*compHost", XrmoptionSepArg, "localhost" },
  { "-s", "*compPort", XrmoptionSepArg, "0" },
};


void start_game(enum who w) {
    int game_ok;
    enum gth_who cgwho = GTH_WHO_NONE;

    switch(w) {
    case WHO_BLACK:
	cgwho = GTH_WHO_BLACK;
	break;
    case WHO_WHITE:
	cgwho = GTH_WHO_WHITE;
	break;
    case WHO_NONE:
	abort();
    }
    game_ok = gth_start_game(cgwho,
				 globalData.compHost,
				 globalData.compServer);
    if (game_ok == -1) {
	fprintf(stderr, "Error connecting to server.\n");
	exit(1);
    }
}

static Pixmap markpixmap(enum who mark) {
  switch (mark) {
  case WHO_BLACK:
    return black_stone_pixmap;
  case WHO_WHITE:
    return white_stone_pixmap;
  case WHO_NONE:
    break;
  }
  return blank_pixmap;
}

void refreshboard(struct board *b) {
  int x, y;

  for (x = 0; x < 5; x++) {
      for (y = 0; y < 5; y++) {
	  int i = x + 5 * y;
	  XtVaSetValues(cellwidget[i],
			XmNlabelPixmap,
			markpixmap(b->cell[x][y]),
			NULL);
      }
  }
}

void resetboard(void) {
  int i;

  for (i = 0; i < 25; i++)
    XtVaSetValues(cellwidget[i],
     XmNlabelPixmap, blank_pixmap,
     NULL);
  XtVaSetValues(status,
   XmNlabelString, emptyString,
   NULL);
}

static void computer_move() {
  enum win winner;
  
  computermove(&currentboard, computerside);
  refreshboard(&currentboard);
  switch (winner = referee(&currentboard)) {
  case WIN_NOTYET:
    break;
  case WIN_BLACK:
  case WIN_WHITE:
    if (whotowin(humanside) == winner)
      XtVaSetValues(status, XmNlabelString, humanWinString, NULL);
    else
      XtVaSetValues(status, XmNlabelString, compWinString, NULL);
    won = 1;
    break;
  case WIN_DRAW:
    XtVaSetValues(status, XmNlabelString, drawString, NULL);
    won = 1;
    break;
  }
}

/*ARGSUSED*/
static void passcallback(Widget w, XtPointer clientData, XtPointer callData) {
  enum gth_state gstatus;
  if (won)
    return;
  currentboard.passes++;
  currentboard.movex = -1;
  currentboard.movey = -1;
  currentboard.movewho = humanside;
  gstatus = gth_make_move(".p");
  if (gstatus == GTH_STATE_ERROR) {
      fprintf(stderr, "Error passing\n");
      abort();
  }
  referee(&currentboard);
  if (won)
    return;
  computer_move();
}

/*ARGSUSED*/
static void movecallback(Widget w, XtPointer clientData, XtPointer callData) {
  int i, x, y;
  enum win winner;
  enum gth_state gstatus;
  char movestr[3];

  if (won)
    return;
  if (currentboard.movewho != computerside)
      return;
  for (i = 0; i < 25; i++)
    if (cellwidget[i] == w)
      break;
  if (i >= 25)
    abort();
  x = i % 5;
  y = i / 5;
  if (!legal_move(&currentboard, x, y, humanside))
    return;
  currentboard.movex = x;
  currentboard.movey = y;
  currentboard.movewho = humanside;
  currentboard.cell[x][y] = humanside;
  captures(&currentboard, x, y, humanside);
  movestr[0] = 'a' + x;
  movestr[1] = '5' - y;
  movestr[2] = '\0';
  refreshboard(&currentboard);
  gstatus = gth_make_move(movestr);
  if (gstatus == GTH_STATE_ERROR) {
      fprintf(stderr, "Error making move on server\n");
      abort();
  }
  switch (winner = referee(&currentboard)) {
  case WIN_NOTYET:
    break;
  case WIN_BLACK:
  case WIN_WHITE:
    if (whotowin(humanside) == winner)
      XtVaSetValues(status, XmNlabelString, humanWinString, NULL);
    else
      XtVaSetValues(status, XmNlabelString, compWinString, NULL);
    gth_end_game();
    won = 1;
    return;
  case WIN_DRAW:
    XtVaSetValues(status, XmNlabelString, drawString, NULL);
    gth_end_game();
    won = 1;
    return;
  }
  computer_move();
}

/*ARGSUSED*/
static void quitcallback(Widget w, XtPointer cld, XtPointer cad) {
  exit(0);
}

/*ARGSUSED*/
static void newcallback(Widget w, XtPointer cld, XtPointer cad) {
  initboard();
}

void displayinit(int *argcp, char **argv) {
  int i;
  Widget shell, layout, board, buttonbox, quit, new, pass;
  static char wname[] = "cellxx";

  shell = XtAppInitialize(&app, "GTHB", options, XtNumber(options),
   argcp, argv, fallback, NULL, 0);
  
  InstallXmCvtBitmapStringToPixmap(shell);

  XtGetApplicationResources(shell, &globalData,
   resources, XtNumber(resources), NULL, 0);

  black_stone_pixmap = globalData.black_stone_pixmap;
  white_stone_pixmap = globalData.white_stone_pixmap;
  blank_pixmap = globalData.blank_pixmap;
  compfirst = globalData.compFirst;
  humanWinString = XmStringCreateLtoR(
   globalData.humanWinString, XmFONTLIST_DEFAULT_TAG );
  compWinString = XmStringCreateLtoR(
   globalData.compWinString, XmFONTLIST_DEFAULT_TAG );
  drawString = XmStringCreateLtoR(
   globalData.drawString, XmFONTLIST_DEFAULT_TAG );
  emptyString = XmStringCreateLtoR(
   "", XmFONTLIST_DEFAULT_TAG );

  layout = XtVaCreateManagedWidget("layout", xmRowColumnWidgetClass,
   shell, NULL);

  board = XtVaCreateManagedWidget("board", xmRowColumnWidgetClass,
   layout, NULL);

  for (i = 0; i < 25; i++) {
    sprintf(&wname[strlen(wname)-2], "%02d", i);   /* XXX template */
    cellwidget[i] =
     XtVaCreateManagedWidget(wname, xmPushButtonWidgetClass, board,
     XmNlabelType, XmPIXMAP,
     XmNlabelPixmap, blank_pixmap,
     NULL);
    XtAddCallback(cellwidget[i], XmNactivateCallback, movecallback, NULL);
  }

  status = XtVaCreateManagedWidget("status", xmLabelWidgetClass, layout, NULL);

  buttonbox = XtVaCreateManagedWidget("buttonBox", xmFormWidgetClass,
   layout, NULL);

  quit = XtVaCreateManagedWidget("quit", xmPushButtonWidgetClass,
   buttonbox, NULL);
  XtAddCallback(quit, XmNactivateCallback, quitcallback, NULL);

  new = XtVaCreateManagedWidget("new", xmPushButtonWidgetClass,
   buttonbox, XmNleftWidget, quit, NULL);
  XtAddCallback(new, XmNactivateCallback, newcallback, NULL);

  pass = XtVaCreateManagedWidget("pass", xmPushButtonWidgetClass,
   buttonbox, XmNleftWidget, new, NULL);
  XtAddCallback(pass, XmNactivateCallback, passcallback, NULL);
  
  XtRealizeWidget(shell);
}

void go(void) {
  XtAppMainLoop(app);
}
